#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include <GL/glut.h>

#define MAXPTS 200
#define EPS 0.000001
#define ANALYTIC 1
#define CASTELJAU 2
#define STEP 0.01

int mousex, mousey;
int sirka=640, vyska=480;
int pts=0;
double P[205][3];
int sel_pt=-1;

int MODE=ANALYTIC;
double comb[205][205];

void display(void) {
  double p1[3],t;
  int i,j;
  double vv[205][205][3];
  
  glClear(GL_COLOR_BUFFER_BIT);
  glColor3d(0.4, 0.4, 0.4);
  glBegin(GL_LINE_STRIP);
    glVertex3dv(P[0]);
    for (i=2;i<pts;i++) glVertex3dv(P[i]);
    glVertex3dv(P[1]);
  glEnd();

  switch (MODE) {
    case ANALYTIC: i=1; 
      glColor3d(0.6, 0.0, 0.0);
      glBegin(GL_LINE_STRIP);
        for (t=0.0;t<1.0+STEP-EPS;t+=STEP) {
          p1[0]=p1[1]=p1[2]=0;
          for (i=2;i<pts;i++) {
            p1[0]+=P[i][0]*comb[pts-1][i-1]*pow(t,i-1)*pow(1-t,pts-i); 
            p1[1]+=P[i][1]*comb[pts-1][i-1]*pow(t,i-1)*pow(1-t,pts-i); 
          }
          p1[0]+=P[0][0]*pow(t,0)*pow(1-t,pts-1); 
          p1[1]+=P[0][1]*pow(t,0)*pow(1-t,pts-1); 
          p1[0]+=P[1][0]*pow(t,pts-1)*pow(1-t,0); 
          p1[1]+=P[1][1]*pow(t,pts-1)*pow(1-t,0); 
          glVertex3dv(p1);
        }
        glVertex3dv(P[1]);
      glEnd();
    break;
    case CASTELJAU:
      glColor3d(0.6, 0.6, 0.0);
      glBegin(GL_LINE_STRIP);
        glVertex3dv(P[1]);
        for (t=STEP;t<1.0-STEP+EPS;t+=STEP) {
          for (i=2;i<pts;i++) { vv[0][i-1][0]=P[i][0]; vv[0][i-1][1]=P[i][1]; vv[0][i-1][2]=P[i][2]; }
          vv[0][0][0]=P[0][0]; vv[0][0][1]=P[0][1]; vv[0][0][2]=P[0][2]; 
          vv[0][pts-1][0]=P[1][0]; vv[0][pts-1][1]=P[1][1]; vv[0][pts-1][2]=P[1][2]; 
          for (i=1;i<pts;i++) for (j=0;j<pts-i;j++) {
            vv[i][j][0]=t*vv[i-1][j][0]+(1-t)*vv[i-1][j+1][0]; 
            vv[i][j][1]=t*vv[i-1][j][1]+(1-t)*vv[i-1][j+1][1]; 
            vv[i][j][2]=t*vv[i-1][j][2]+(1-t)*vv[i-1][j+1][2]; 
          }
          glVertex3dv(vv[pts-1][0]);
        }
        glVertex3dv(P[0]);
      glEnd();
    break;
  }
  glPointSize(10);
  glColor3d(0.6, 0.0, 0.0);
  for (i=2;i<pts;i++) { glBegin(GL_POINTS); glVertex3dv(P[i]); glEnd(); }
  glColor3d(0.6, 0.6, 0.0);
  glBegin(GL_POINTS); glVertex3dv(P[0]); glVertex3dv(P[1]); glEnd();

  glFlush();
}

void reshape(int w, int h) {
  sirka=w; vyska=h;
  glViewport(0, 0, w, h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0, w, 0, h);
  glScalef(1, -1, 1);
  glTranslatef(0, -h, 0);
  glMatrixMode(GL_MODELVIEW);
}

void keyPressed(unsigned char key, int x, int y) { 
  if (key == 27) exit(0); 
}

void motion(int x, int y) {
  if (sel_pt!=-1) { P[sel_pt][0]=x; P[sel_pt][1]=y; }
  mousex = x; mousey = y;
  glutPostRedisplay();
}

void mouse(int b, int s, int x, int y) {
  int i,j,ok;

  mousex = x; mousey = y;
  if (b==GLUT_LEFT_BUTTON){
    if (s==GLUT_DOWN) {
      ok=1;
      for (i=0;i<pts;i++) if ((abs(P[i][0]-x)<=4) && (abs(P[i][1]-y)<=4)) {
        ok=0; if (i>1) { 
          for (j=i;j<pts-1;j++) { P[j][0]=P[j+1][0]; P[j][1]=P[j+1][1]; P[j][2]=P[j+1][2]; }
          pts--; 
        } 
        break;
      }
      if ((ok) && (pts<MAXPTS)) { P[pts][0]=x; P[pts][1]=y; P[pts][2]=0; pts++; }
    }
  } else {
    if (s==GLUT_DOWN) {
      sel_pt=-1;
      for (i=0;i<pts;i++) if ((abs(P[i][0]-x)<=4) && (abs(P[i][1]-y)<=4)) { sel_pt=i; break; }
    } else { sel_pt=-1; }
  }
  
  glutPostRedisplay();
  motion(x,y);
}

void menu(int choice) {
  switch(choice) {
    case 1: MODE=ANALYTIC; break;
    case 2: MODE=CASTELJAU; break;
  }
  glutPostRedisplay();
}

void PreComputeComb(void){
  int i,j;
  for (i=0;i<=200;i++) for (j=0;j<=i;j++) {
    if ((j==0)||(j==i)) { comb[i][j]=1.0; }
      else comb[i][j]=comb[i-1][j-1]+comb[i-1][j];
  }
}

int main(int argc, char **argv) {
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_SINGLE | GLUT_RGB);
  
  glutInitWindowSize(640, 480); glutInitWindowPosition(0, 0);
  glutCreateWindow("Miso Forisek -- DU #2");
  glClearColor(0.0, 0.0, 0.0, 0.0); glColor3d(1.0, 0.0, 0.0); glLineWidth(1.0);
  
  glutDisplayFunc(display);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(keyPressed);
  glutMotionFunc(motion);
  glutMouseFunc(mouse);
  
  glutCreateMenu(menu); glutAddMenuEntry("Analytic", 1); glutAddMenuEntry("Casteljau", 2);
  glutAttachMenu(GLUT_RIGHT_BUTTON);

  pts=2;
  P[0][0]=140.0; P[0][1]=240.0; P[0][2]=0.0;
  P[1][0]=500.0; P[1][1]=240.0; P[1][2]=0.0;
  PreComputeComb();
          
  glutMainLoop();
  return 0;
}
